// (C) 1992-2015 Altera Corporation. All rights reserved.                         
// Your use of Altera Corporation's design tools, logic functions and other       
// software and tools, and its AMPP partner logic functions, and any output       
// files any of the foregoing (including device programming or simulation         
// files), and any associated documentation or information are expressly subject  
// to the terms and conditions of the Altera Program License Subscription         
// Agreement, Altera MegaCore Function License Agreement, or other applicable     
// license agreement, including, without limitation, that your use is for the     
// sole purpose of programming logic devices manufactured by Altera and sold by   
// Altera or its authorized distributors.  Please refer to the applicable         
// agreement for further details.                                                 
    


/* ===- acl_pcie_config.cpp  ------------------------------------------ C++ -*-=== */
/*                                                                                 */
/*                         Altera OpenCL MMD Driver                                */
/*                                                                                 */
/* ===-------------------------------------------------------------------------=== */
/*                                                                                 */
/* This file implements the class to handle functions that program the FPGA.       */
/* The declaration of the class lives in the acl_pcie_config.h.                    */
/*                                                                                 */
/* ===-------------------------------------------------------------------------=== */


// common and its own header files
#include "acl_pcie.h"
#include "acl_pcie_config.h"

// other header files inside MMD driver
#include "acl_pcie_debug.h"
#include "hw_pcie_cvp_constants.h"

// other standard header files
#include <stdlib.h>
#include <string.h>

#if defined(LINUX)
#  include <unistd.h>
#endif   // LINUX



// Function to install the signal handler for Ctrl-C
// Implemented inside acl_pcie.cpp
extern int install_ctrl_c_handler(int ingore_sig);



ACL_PCIE_CONFIG::ACL_PCIE_CONFIG(WDC_DEVICE_HANDLE device)
{
   m_device = device;   

#if defined(WINDOWS)
   m_slot   = ((WDC_DEVICE *)device)->slot.pciSlot;
   ACL_PCIE_ASSERT( find_upstream_slot(m_slot, &m_upstream), "cannot find the upstream slot.\n" ); 
#endif   // WINDOWS
   
   return;
}

ACL_PCIE_CONFIG::~ACL_PCIE_CONFIG() 
{
}



// Change the core only via PCIe, using an in-memory image of the core.rbf
// For Linux, the actual implementation of CvP is inside the kernel mode driver.
// Return 0 on success.
int ACL_PCIE_CONFIG::program_core_with_CvP_image(int *core_bitstream, size_t core_rbf_len)
{
   int cvp_result = 1;   // set to default - failure 

   ACL_PCIE_ERROR_IF( core_bitstream == NULL, return 1, 
      "core_bitstream is an NULL pointer.\n" );
   ACL_PCIE_ERROR_IF( core_rbf_len < 1000000, return 1, 
      "size of core rbf file is suspiciously small.\n" );

#if defined(WINDOWS)
   int cvp_failed = 1;   // set to default - failure 
   unsigned char status;

   ACL_PCIE_ERROR_IF( !read_bit(m_device, CVP_EN), return 1, 
      "fail reading bit CVP_EN for CvP.\n");
   
   // Setup
   write_bit(m_device, HIP_CLK_SEL, 1);
   write_bit(m_device, CVP_MODE,    1);
   Sleep(2);
   switch_clock(m_device);

   write_bit(m_device, CVP_CONFIG,  1);
   Sleep(2);
   switch_clock(m_device);

   status = wait_for_bit(m_device, CVP_CONFIG_READY, 1);
   ACL_PCIE_ERROR_IF( !status, goto teardown, 
      "timed out waiting for CVP_CONFIG_READY to become 1!\n");

   switch_clock(m_device);
   write_bit(m_device, START_XFER, 1);

   status  = wait_for_bit(m_device, HIP_CLK_SEL, 1); 
   status &= wait_for_bit(m_device, CVP_MODE, 1);
   ACL_PCIE_ERROR_IF( !status, goto teardown, 
      "timed out waiting for HIP_CLK_SEL and CVP_MODE to be 1!\n");

   // Transfer
   ACL_PCIE_DEBUG_MSG(":: Setup is done. Starting to write CvP data!\n");
   
   prepare_for_pgm_data(m_device);
   if( send_pgm_data(m_device, (unsigned int *)core_bitstream, core_rbf_len) == core_rbf_len){
      ACL_PCIE_DEBUG_MSG(":: Reached the end of the core programming file.\n");
   }

   // Teardown
teardown:
   write_bit(m_device, START_XFER, 0);
   write_bit(m_device, CVP_CONFIG, 0);
   switch_clock(m_device);
   
   // wait for the CB to say it's done with CvP
   status = wait_for_bit(m_device, CVP_CONFIG_READY, 0);
   ACL_PCIE_ERROR_IF( !status, /* do nothing */, 
      "timed out waiting for CVP_CONFIG_READY to become 0!\n");

   if( read_bit(m_device, CVP_CFG_ERR_LATCH) ){
      ACL_PCIE_ERROR_IF(1, /* do nothing */, "configuration error detected!\n");
      // write a 1 to clear the config space error bit      
      write_bit(m_device, CVP_CFG_ERR_LATCH, 1);          
   } else {
      cvp_failed = 0;
   }

   write_bit(m_device, CVP_MODE,    0);
   write_bit(m_device, HIP_CLK_SEL, 0);
   
   if( !cvp_failed ){
      // wait for the Application Layer to be ready for normal operation
      if( !wait_for_bit(m_device, PLD_CLK_IN_USE, 1) ){
         ACL_PCIE_ERROR_IF(1, /* do nothing */, "timed out waiting for PLD_CLK_IN_USE to become 1!\n");
      } else if( !wait_for_bit(m_device, USER_MODE, 1) ){
         ACL_PCIE_ERROR_IF(1, /* do nothing */, "timed out waiting for USER_MODE to become 1!\n");
      } else {
         Sleep(200);
         cvp_result = 0;
      }
   }
#endif   // WINDOWS
#if defined(LINUX)
   struct acl_cmd cmd_cvp = { ACLPCI_CMD_BAR, ACLPCI_CMD_DO_CVP, NULL, NULL };

   cmd_cvp.user_addr = core_bitstream;
   cmd_cvp.size      = core_rbf_len;
     
   cvp_result = read( m_device, &cmd_cvp, sizeof(cmd_cvp) );

#endif   // LINUX

   return cvp_result;   
}

// Program using an in-memory image of RBF (Not core rbf) file
// Return 0 on success.
int ACL_PCIE_CONFIG::program_with_RBF_image(char *rbf_image, size_t rbf_len) 
{
   int program_failed = 1;
   
   ACL_PCIE_ERROR_IF(rbf_image == NULL, return 1, 
      "rbf_image is a NULL pointer.\n");
      
#if defined(WINDOWS)
   ACL_PCIE_INFO("RBF from memory is not implemented on Windows\n");
#endif   // WINDOWS
#if defined(LINUX)
   // write image to temp file
   char tmp_filename[FILENAME_MAX];
   char buf[FILENAME_MAX];
   tmpnam(tmp_filename);
   FILE *tmp = fopen (tmp_filename, "wb");
   
   ACL_PCIE_ERROR_IF( tmp == NULL, return 1, 
      "couldn't open tmp file %s for writing!\n", tmp_filename);

   fwrite (rbf_image, sizeof(char), rbf_len, tmp);
   fclose (tmp);
   
   sprintf (buf, "cat %s > /dev/fpga0", tmp_filename);
   system (buf);
   program_failed = (fpga_in_user_mode() == 0); // failed if NOT in user mode

   ACL_PCIE_ERROR_IF( remove (tmp_filename) != 0, /* do nothing */, 
      "couldn't delete temporary RBF file %s\n", tmp_filename);
#endif   // LINUX

   return program_failed;
}

// Program the FPGA using a given SOF file
// Quartus is needed for this, because, 
//   quartus_pgm is used to progrma the board through USB blaster
// For Linux, when the kernel driver is asked to save/load_pcie_control_regs(),
//   it will also disable/enable the aer on the upstream, so no need to 
//   implement those here.
// Return 0 on success.
int ACL_PCIE_CONFIG::program_with_SOF_file(const char *filename) 
{
   const int MAX_ATTEMPTS   = 3;
   int       program_failed = 1;

   char cmd[4*1024];
   sprintf(cmd, "quartus_pgm -c 1 -m jtag -o \"P;%s\"", filename);
   ACL_PCIE_INFO("executing \"%s\"\n", cmd);

#if defined(WINDOWS)
   const DWORD PCIE_AER_CAPABILITY_ID                 = 0x0001;
   const DWORD PCIE_AER_UNCORRECTABLE_STATUS_OFFSET   = 0x4;
   const DWORD PCIE_AER_UNCORRECTABLE_MASK_OFFSET     = 0x8;
   const DWORD PCIE_AER_CORRECTABLE_STATUS_OFFSET     = 0x10;
   const DWORD PCIE_AER_SURPRISE_DOWN_BIT             = (1<<5);

   DWORD WD_status;
   int   status;

   // Find location in extended config space of AER (id 0x0001) in upstream port
   DWORD location_of_aer;
   int has_aer = find_extended_capability( m_upstream, PCIE_AER_CAPABILITY_ID, &location_of_aer);

   // Disable AER by setting bit 5 (surprise down) in uncorrectable error mask register (offset 0x8, 4 bytes)
   // This prevents the root complex from reporting the error further upstream
   UINT32 data;
   if (has_aer) {
      WD_status  = WDC_PciReadCfgBySlot (&m_upstream, location_of_aer+PCIE_AER_UNCORRECTABLE_MASK_OFFSET, &data, 4);
      data      |= PCIE_AER_SURPRISE_DOWN_BIT;
      WD_status |= WDC_PciWriteCfgBySlot(&m_upstream, location_of_aer+PCIE_AER_UNCORRECTABLE_MASK_OFFSET, &data, 4);

      ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
         "WDC_PciReadCfgBySlot/WDC_PciWriteCfgBySlot failed when programming with SOF file.\n");
   }
#endif   // WINDOWS

   // Set the program to ignore the ctrl-c signal
   // This setting will be inherited by the systme() function call below,
   // so that the quartus_pgm call won't be interrupt by the ctrl-c signal.
   install_ctrl_c_handler(1 /* ignore the signal */);

   // program the FPGA by executing the command
   for( int attempts = 0; attempts < MAX_ATTEMPTS && program_failed; attempts++ ){
      program_failed = system(cmd);
#if defined(WINDOWS)
      WDC_Sleep( 2000000, WDC_SLEEP_NON_BUSY );
#endif   // WINDOWS
#if defined(LINUX)
      sleep(2);
#endif   // LINUX
   }

   // Restore the original custom ctrl-c signal handler
   install_ctrl_c_handler(0 /* use the custom signal handler */);

#if defined(WINDOWS)
   status = load_pci_control_regs();
   ACL_PCIE_ERROR_IF(status, return -1, 
      "load_pci_control_regs failed programming with SOF file.\n");

   // Restore AER by resetting bit 5 (surprise down) in uncorrectable error mask register (offset 0x8)
   // But first, need to clear the surprise down error (bit 5) in the status register (offsets 0x4 and 0x10)
   // Errors are cleared by writing a 1 to that bit, so to clear all errors, write back the value just read
   if (has_aer) {
      WD_status  = WDC_PciReadCfgBySlot ( &m_upstream, location_of_aer+PCIE_AER_UNCORRECTABLE_STATUS_OFFSET, &data, 4 );
      WD_status |= WDC_PciWriteCfgBySlot( &m_upstream, location_of_aer+PCIE_AER_UNCORRECTABLE_STATUS_OFFSET, &data, 4 );

      WD_status |= WDC_PciReadCfgBySlot ( &m_upstream, location_of_aer+PCIE_AER_CORRECTABLE_STATUS_OFFSET, &data, 4 );
      WD_status |= WDC_PciWriteCfgBySlot( &m_upstream, location_of_aer+PCIE_AER_CORRECTABLE_STATUS_OFFSET, &data, 4 );

      WD_status |= WDC_PciReadCfgBySlot ( &m_upstream, location_of_aer+PCIE_AER_UNCORRECTABLE_MASK_OFFSET, &data, 4 );
      data      &= ~PCIE_AER_SURPRISE_DOWN_BIT;
      WD_status |= WDC_PciWriteCfgBySlot( &m_upstream, location_of_aer+PCIE_AER_UNCORRECTABLE_MASK_OFFSET, &data, 4 );

      ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
         "WDC_PciReadCfgBySlot/WDC_PciWriteCfgBySlot failed when programming with SOF file.\n");
   }

   // Retrain the link after reprogramming the FPGA
   status = retrain_link(m_slot);
   ACL_PCIE_ERROR_IF(status, return -1, 
      "fail retrain the link after programming with SOF file.\n");
#endif   // WINDOWS

   return program_failed;
}



// Functions to save/load control registers form PCI Configuration Space
// This saved registers are used to restore the PCIe link after reprogramming
// through methods other than CvP
// For Windows, the register values are stored in this class, and do 
//   nothing else
// For Linux, the register values are stored inside the kernel driver, 
//   And, it will disable the interrupt and the aer on the upstream, 
//   when the save_pci_control_regs() function is called. They will 
//   be enable when load_pci_control_regs() is called.  
// Return 0 on success
int ACL_PCIE_CONFIG::save_pci_control_regs() 
{
   int save_failed = 1;
   
#if defined(WINDOWS)
   save_failed = (int)WDC_PciReadCfg(m_device, 0, m_config_space, CONFIG_SPACE_SIZE);
#endif   // WINDOWS
#if defined(LINUX)
   struct acl_cmd cmd_save = { ACLPCI_CMD_BAR, ACLPCI_CMD_SAVE_PCI_CONTROL_REGS, NULL, NULL };
   save_failed = read(m_device, &cmd_save, 0);
   disable_bridges();
#endif   // LINUX

   return save_failed;
}

int ACL_PCIE_CONFIG::load_pci_control_regs() 
{
   int load_failed = 1;
   
#if defined(WINDOWS)
   load_failed = (int)WDC_PciWriteCfg(m_device, 0, m_config_space, CONFIG_SPACE_SIZE);
#endif   // WINDOWS
#if defined(LINUX)
   struct acl_cmd cmd_load = { ACLPCI_CMD_BAR, ACLPCI_CMD_LOAD_PCI_CONTROL_REGS, NULL, NULL };
   enable_bridges();
   load_failed = read(m_device, &cmd_load, 0);
#endif   // LINUX

   return load_failed;
}



// Functions to query the PCI related information
// Use NULL as input for the info that you don't care about
// Return 0 on success.
int ACL_PCIE_CONFIG::query_pcie_info(unsigned int *pcie_gen, unsigned int *pcie_num_lanes, char *pcie_slot_info_str)
{
   int            status = 0;
#if defined(WINDOWS)
   const DWORD  LINK_STATUS_OFFSET = 0x12;
   DWORD        board_caps_offset;
   unsigned int read_value = 0;

   // Find the PCIe Capabilities Structure offset
   ACL_PCIE_ERROR_IF( find_capabilities(m_slot, &board_caps_offset) != 1, return -1, 
      "cannot find capabilities when querying PCIE info.\n");

   DWORD board_status = board_caps_offset + LINK_STATUS_OFFSET;
   DWORD WD_status = WDC_PciReadCfgBySlot( &m_slot, board_status, &read_value, 2 );
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when querying PCIE info.\n");

   if( pcie_gen != NULL) {
      *pcie_gen = read_value & 0xf;
   }

   if( pcie_num_lanes != NULL ) {
      *pcie_num_lanes = (read_value & 0xf0) >> 4;
   }

   if( pcie_slot_info_str != NULL ) {
      sprintf(pcie_slot_info_str, "%02x:%02x.%02x", 
         m_slot.dwBus, m_slot.dwSlot, m_slot.dwFunction);
   }
#endif   // WINDOWS
#if defined(LINUX)
   struct acl_cmd driver_cmd;

   if( pcie_gen != NULL ) {
      driver_cmd.bar_id      = ACLPCI_CMD_BAR;
      driver_cmd.command     = ACLPCI_CMD_GET_PCI_GEN;
      driver_cmd.device_addr = NULL;
      driver_cmd.user_addr   = pcie_gen;
      driver_cmd.size        = sizeof(*pcie_gen);
      status |= read (m_device, &driver_cmd, sizeof(driver_cmd));
   }

   if( pcie_num_lanes != NULL ) {
      driver_cmd.bar_id      = ACLPCI_CMD_BAR;
      driver_cmd.command     = ACLPCI_CMD_GET_PCI_NUM_LANES;
      driver_cmd.device_addr = NULL;
      driver_cmd.user_addr   = pcie_num_lanes;
      driver_cmd.size        = sizeof(*pcie_num_lanes);
      status |= read (m_device, &driver_cmd, sizeof(driver_cmd));
   }

   if( pcie_slot_info_str != NULL ) {
      driver_cmd.bar_id      = ACLPCI_CMD_BAR;
      driver_cmd.command     = ACLPCI_CMD_GET_PCI_SLOT_INFO;
      driver_cmd.device_addr = NULL;
      driver_cmd.user_addr   = pcie_slot_info_str;
      driver_cmd.size        = sizeof(pcie_slot_info_str);
      status |= read (m_device, &driver_cmd, sizeof(driver_cmd));
   }
#endif   // LINUX

   return status;
}



// Disable communication bridges between ARM and FPGA
void ACL_PCIE_CONFIG::disable_bridges() 
{
#if defined(ARM)
   system ("echo 0 > /sys/class/fpga-bridge/fpga2hps/enable");
   system ("echo 0 > /sys/class/fpga-bridge/hps2fpga/enable");
   system ("echo 0 > /sys/class/fpga-bridge/lwhps2fpga/enable");
#endif   // ARM
}

// Enable communication bridges between ARM and FPGA
void ACL_PCIE_CONFIG::enable_bridges() 
{
#if defined(ARM)
   system ("echo 1 > /sys/class/fpga-bridge/fpga2hps/enable");
   system ("echo 1 > /sys/class/fpga-bridge/hps2fpga/enable");
   system ("echo 1 > /sys/class/fpga-bridge/lwhps2fpga/enable");
#endif   // ARM
}

int ACL_PCIE_CONFIG::fpga_in_user_mode() 
{
  #if defined(ARM)
  
    #define BUF_SIZE 1024
    char buf[BUF_SIZE];
    const char *status_file = "/sys/class/fpga/fpga0/status";
    char *fgets_res = NULL;  
    
    FILE *status = fopen (status_file, "r");
    if (status == NULL) {
      fprintf (stderr, "Couldn't open FPGA status from %s!\n", status_file);
      return 0;
    }
    
    fgets_res = fgets (buf, BUF_SIZE, status);
    fclose (status);
      
    if (fgets_res == NULL) {
      fprintf (stderr, "Couldn't read FPGA status from %s!\n", status_file);
      return 0;
    }
      
    if (strstr (buf, "user mode") == NULL) {
      fprintf (stderr, "After reprogramming, FPGA is not in user mode (%s)!\n", buf);
      return 0;
    } 
    
    // If here, FPGA is in user mode
    return 1;
  #else
    // Assuming that for non-SoC devices, the fpga is always in user mode
    return 1;
  #endif
}



#if defined(WINDOWS)
// Retrain the PCIe link after programming FPGA with SOF file (Windows only)
// Return 0 on success
int ACL_PCIE_CONFIG::retrain_link(WD_PCI_SLOT &slot) 
{
   const int DESIRED_WIDTH = 8;
   const int GEN3_SPEED    = 3, GEN2_SPEED    = 2, GEN1_SPEED    = 1;

   const DWORD LINK_STATUS_OFFSET       = 0x12;
   const DWORD LINK_CONTROL_OFFSET      = 0x10;
   const DWORD LINK_CAPABILITIES_OFFSET = 0x0c;
   const DWORD RETRAIN_LINK_BIT         = (1 << 5);
   const DWORD TRAINING_IN_PROGRESS_BIT = (1 << 11);

   DWORD        WD_status;
   DWORD        board_caps_offset, usa_caps_offset;

   // Find the PCIe Capabilities Structure offset
   ACL_PCIE_ERROR_IF( find_capabilities(slot, &board_caps_offset) != 1, return -1, 
      "cannot find capabilities when retraining the link.\n");

   // Find the PCIe Capabilities Structure offset for the upstream port
   ACL_PCIE_ERROR_IF( find_capabilities(m_upstream, &usa_caps_offset)!= 1, return -1, 
      "cannot find PCIe capabilities structure for upstream port.\n" );

   // variables for checking status and link caps
   unsigned int status;
   UINT32       link_caps;

   DWORD board_status    = board_caps_offset + LINK_STATUS_OFFSET; 
   DWORD board_link_caps = board_caps_offset + LINK_CAPABILITIES_OFFSET;
   DWORD usa_status      = usa_caps_offset   + LINK_STATUS_OFFSET;
   DWORD usa_control     = usa_caps_offset   + LINK_CONTROL_OFFSET;
   DWORD usa_link_caps   = usa_caps_offset   + LINK_CAPABILITIES_OFFSET;

   // Check the current status for the board
   WD_status = WDC_PciReadCfgBySlot( &slot, board_status, &status, 2);
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when retraining the link.\n");

   if((status & 0xf) == GEN3_SPEED){ 
      ACL_PCIE_INFO("Link currently operating at 8 GT/s.\n");
   } else if((status & 0xf) == GEN2_SPEED){
      ACL_PCIE_INFO("Link currently operating at 5 GT/s.\n");
   } else if((status & 0xf) == GEN1_SPEED){
      ACL_PCIE_INFO("Link currently operating at 2.5 GT/s\n");
   } else{
      ACL_PCIE_INFO("Link currently operating at an unknown speed. Status register is %x\n", status);
   }

   // check the current link caps for the board
   WD_status = WDC_PciReadCfgBySlot( &slot, board_link_caps, &link_caps, 4 );
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when retraining the link.\n");

   unsigned int board_max_width       = (link_caps >> 4) & 0x3f;   // Link width: bits 9:4
   unsigned int supported_speed_bits  = (link_caps & 0x0f);        // Link speed: bits 3:0
   unsigned int desired_trained_speed = supported_speed_bits;

   float board_max_speed = (supported_speed_bits == GEN1_SPEED) ? 2.5f :
                           (supported_speed_bits == GEN2_SPEED) ? 5.0f :
                           (supported_speed_bits == GEN3_SPEED) ? 8.0f : //gen 3
                                                                  0.0f ;
   ACL_PCIE_DEBUG_MSG("Board max upstream width: x%d\n",       board_max_width);
   ACL_PCIE_DEBUG_MSG("Board max upstream speed: %.1f GT/s\n", board_max_speed);

   // check the current link caps for the upstream port
   WD_status = WDC_PciReadCfgBySlot( &m_upstream, usa_link_caps, &link_caps, 4 );
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when retraining the link.\n");

   unsigned int max_width = (link_caps >> 4) & 0x3f;   // Link width: bits 9:4
   supported_speed_bits   = (link_caps & 0x0f);        // Link speed: bits 3:0
   desired_trained_speed  = supported_speed_bits < desired_trained_speed ? 
      supported_speed_bits : desired_trained_speed;

   float max_speed = (supported_speed_bits == GEN1_SPEED) ? 2.5f :
                     (supported_speed_bits == GEN2_SPEED) ? 5.0f :
                     (supported_speed_bits == GEN3_SPEED) ? 8.0f : //gen 3
                                                            0.0f ;
   ACL_PCIE_DEBUG_MSG("Max upstream width: x%d\n",       max_width);
   ACL_PCIE_DEBUG_MSG("Max upstream speed: %.1f GT/s\n", max_speed);

   if(max_width < DESIRED_WIDTH) 
      ACL_PCIE_WARN_MSG("PCIe slot does not support more than %d lanes, PCIe throughput will be adversely affected.\n", DESIRED_WIDTH);
   if(supported_speed_bits < 2)
      ACL_PCIE_WARN_MSG("PCIe slot does not support gen2 operation.  PCIe throughput will be adversely affected.\n");

   // Retrain the link
   UINT16       control;
   WD_status  = WDC_PciReadCfgBySlot ( &m_upstream, usa_status,  &status,  2 );
   WD_status |= WDC_PciReadCfgBySlot ( &m_upstream, usa_control, &control, 2 );
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when retraining the link.\n");

   ACL_PCIE_DEBUG_MSG("Upstream status before retrain: 0x%x\n",  status);
   ACL_PCIE_DEBUG_MSG("Upstream control before retrain: 0x%x\n", control);

   DWORD wdata = control | RETRAIN_LINK_BIT;
   WD_status = WDC_PciWriteCfgBySlot( &m_upstream, usa_control, &wdata, 2 );
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciWriteCfgBySlot failed when retraining the link.\n");

   // Wait for training to complete
   bool training = true;
   UINT32 timeout = 0;
   while(training && timeout < 50) {
      WD_status |= WDC_PciReadCfgBySlot( &m_upstream, usa_status, &status, 2 );
      training = (status & (TRAINING_IN_PROGRESS_BIT));
      WDC_Sleep( 1000, WDC_SLEEP_BUSY ); // 1 ms
      ++timeout;
   }

   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when retraining the link.\n");

   ACL_PCIE_ERROR_IF(training, return -1, 
      "Link training timed out, PCIe link not established. \n");

   ACL_PCIE_DEBUG_MSG("Link training completed in %d ms.\n", timeout);

   // Check the new link speed
   WD_status = WDC_PciReadCfgBySlot( &slot, board_status, &status, 2 );
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return -1, 
      "WDC_PciReadCfgBySlot failed when retraining the link.\n");

   unsigned int lanes         = (status & 0xf0) >> 4;
   unsigned int trained_speed = status & 0xf;
   unsigned int real_trained_speed_per_lane = (trained_speed==GEN1_SPEED) ? 250U : 
                                              (trained_speed==GEN2_SPEED) ? 500U : 
                                              (trained_speed==GEN3_SPEED) ? 1000U : 
                                                                            0U;
   ACL_PCIE_INFO("Link operating at Gen %u with %u lanes.\n", trained_speed, lanes);
   ACL_PCIE_INFO("Expected peak bandwidth = %u MB/s\n",lanes*real_trained_speed_per_lane);

   unsigned int achievable_max_width = (board_max_width < max_width) ? 
      board_max_width : max_width;   //max_width is that of upstream

   if (lanes < achievable_max_width)
      ACL_PCIE_WARN_MSG("Link trained %d lanes whereas %d were desired.\n", lanes, achievable_max_width);

   ACL_PCIE_ERROR_IF(trained_speed < desired_trained_speed, return -1, 
      "Link trained at gen %d whereas gen %d was desired.\n", trained_speed, desired_trained_speed);

   return 0;  // success
}



// Try to find the upstream port for the current board slot
// Return 1 when the target is found
int ACL_PCIE_CONFIG::find_upstream_slot(WD_PCI_SLOT &slot, WD_PCI_SLOT *upstream) 
{
   int    found = 0;
   UINT32 data;
   DWORD  WD_status;

   // Enumerate all devices
   WDC_PCI_SCAN_RESULT scan_result;
   WD_status = WDC_PciScanDevices(0, 0, &scan_result);
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return 0, 
      "WDC_PciScanDevices failed when finding upstream slot.\n");

   for(unsigned int i = 0; i < scan_result.dwNumDevices; i++) {
      // Looking for a type 1 header
      // The top bit indicates multi-function support, ignore it (mask with 0x7f, not 0xff)
      WD_status = WDC_PciReadCfgBySlot( &(scan_result.deviceSlot[i]), 0x0C, &data, 4 );
      if ( WD_status != WD_STATUS_SUCCESS || ((data >> 16) & 0x7f) != 1 ) {
         continue;
      }

      // Check the secondary bus number - looking for the source of the bus with ID [bus]
      WD_status = WDC_PciReadCfgBySlot( &(scan_result.deviceSlot[i]), 0x18, &data, 4 );
      if ( WD_status != WD_STATUS_SUCCESS || ((data >> 8) & 0xff) != slot.dwBus ) {
         continue;
      }

      // Found the source of this bus - verify there is only one
      ACL_PCIE_ERROR_IF(found, return 0, 
         "Found multiple sources(%d and %d) for bus %d!\n", 
         upstream->dwBus, scan_result.deviceSlot[i], slot.dwBus);

      found = 1;
      *upstream = scan_result.deviceSlot[i];
   }

   return found;
}

// Try to find the PCIe capabilities structure offset
// Return 1 when the target is found
int ACL_PCIE_CONFIG::find_capabilities(WD_PCI_SLOT &slot, DWORD *offset)
{
   int    found = 0;
   UINT32 data;
   DWORD  WD_status;

   // Read the head of the caps list
   WD_status = WDC_PciReadCfgBySlot(&slot, 0x34, &data, 4);
   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return 0, 
      "WDC_PciReadCfgBySlot failed when finding capabilities.\n");
   ACL_PCIE_ERROR_IF(data == 0xffffffff, return 0,
      "get 0xffffffff when finding capabilities.\n");

   DWORD cap_ptr = (data & 0xff);

   // Walk the list of caps to find the PCIe caps structure
   while(cap_ptr != 0 && !found) {
      WD_status |= WDC_PciReadCfgBySlot(&slot, cap_ptr, &data, 4);

      // Grab the current caps struct and link to the next one
      *offset = cap_ptr;
      cap_ptr = ((data >> 8) & 0xff);

      // Looking for the PCIe caps ID
      if( (data & 0xff) != 0x10 )
         continue;

      // Found it :)
      found = 1;
   }

   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return 0, 
      "WDC_PciReadCfgBySlot failed when finding capabilities.\n");

   return found;
}   

// Try to find the location in extended config space of the given ID
// Return 1 when the target is found
int ACL_PCIE_CONFIG::find_extended_capability(WD_PCI_SLOT &slot, UINT16 ex_cap_id_to_find, DWORD *offset) 
{
   int found = 0;
   UINT32 data;
   DWORD WD_status = WD_STATUS_SUCCESS;

   DWORD ex_cap_ptr = 0x100;

   // Walk the link list of capabilities
   while(ex_cap_ptr != 0 && !found) {
      WD_status |= WDC_PciReadCfgBySlot( &slot, ex_cap_ptr, &data, 4 );

      // Grab the current caps struct and link to the next one
      *offset = ex_cap_ptr;
      ex_cap_ptr = ((data >> 20) & 0xfff);	//next extended capability on bits 31:20

      // extended capability id is on bottom 16 bits
      if( (data & 0xffff) != ex_cap_id_to_find )
         continue;

      // Found it :)
      found = 1;
   }

   ACL_PCIE_ERROR_IF(WD_status != WD_STATUS_SUCCESS, return 0, 
      "WDC_PciReadCfgBySlot failed when finding extended capability.\n");

   return found;
}
#endif   // WINDOWS

   
#if defined(WINDOWS)
void ACL_PCIE_CONFIG::get_mask_and_offset(unsigned char whichBit, unsigned char *bitMask, unsigned char *bitRegOffset)
{
   switch (whichBit) {
      case DATA_ENCRYPTED:
         *bitRegOffset = OFFSET_CVP_STATUS;
         *bitMask = MASK_DATA_ENCRYPTED;
         break;
      case DATA_COMPRESSED:
         *bitRegOffset = OFFSET_CVP_STATUS;
         *bitMask = MASK_DATA_COMPRESSED;
         break;
      case CVP_CONFIG_READY:
         *bitRegOffset = OFFSET_CVP_STATUS;
         *bitMask = MASK_CVP_CONFIG_READY;
         break;
      case CVP_CONFIG_ERROR:
         *bitRegOffset = OFFSET_CVP_STATUS;
         *bitMask = MASK_CVP_CONFIG_ERROR;
         break;
      case CVP_EN:
         *bitRegOffset = OFFSET_CVP_STATUS;
         *bitMask = MASK_CVP_EN;
         break;
      case USER_MODE:
         *bitRegOffset = OFFSET_CVP_STATUS;
         *bitMask =  MASK_USER_MODE;
         break;
      case PLD_CLK_IN_USE:
         *bitRegOffset = OFFSET_CVP_STATUS+1;
         *bitMask = MASK_PLD_CLK_IN_USE;
         break;
      case CVP_MODE:
         *bitRegOffset = OFFSET_CVP_MODE_CTRL;
         *bitMask = MASK_CVP_MODE;
         break;
      case HIP_CLK_SEL:
         *bitRegOffset = OFFSET_CVP_MODE_CTRL;
         *bitMask = MASK_HIP_CLK_SEL;
         break;
      case CVP_CONFIG:
         *bitRegOffset = OFFSET_CVP_PROG_CTRL;
         *bitMask = MASK_CVP_CONFIG;
         break;
      case START_XFER:
         *bitRegOffset = OFFSET_CVP_PROG_CTRL;
         *bitMask = MASK_START_XFER;
         break;
      case CVP_CFG_ERR_LATCH:
         *bitRegOffset = OFFSET_UNC_IE_STATUS;
         *bitMask = MASK_CVP_CFG_ERR_LATCH;
         break;
      default:
         *bitRegOffset = 0xff;
         *bitMask = 0xff;
         break;
   }
}

unsigned char ACL_PCIE_CONFIG::wait_for_bit(WDC_DEVICE_HANDLE device, unsigned char whichBit, unsigned char bitValue)
{
   int max_num_tries = 10;
   int itry = 0;
   while (read_bit(device, whichBit) != bitValue && itry < max_num_tries){
      itry++;
      Sleep(1);   /* in milliseconds */
   }
   return (itry < max_num_tries) ? 1U : 0U;
}

unsigned char ACL_PCIE_CONFIG::read_bit(WDC_DEVICE_HANDLE device, unsigned char whichBit) 
{
   unsigned char bitMask, bitRegOffset;
   BYTE byteRead;
   get_mask_and_offset(whichBit, &bitMask, &bitRegOffset);
   WDC_PciReadCfg8(device, (DWORD)OFFSET_VSEC+bitRegOffset, &byteRead);
   return (byteRead & bitMask) ? 1U : 0U; 
}

void ACL_PCIE_CONFIG::write_bit(WDC_DEVICE_HANDLE device, unsigned char whichBit, unsigned char bitValue) 
{
   unsigned char bitMask, bitRegOffset;
   BYTE byteValue;   
   switch (whichBit) {
      case CVP_MODE:
      case HIP_CLK_SEL:
      case CVP_CONFIG:
      case START_XFER:
      case CVP_CFG_ERR_LATCH:
         get_mask_and_offset(whichBit, &bitMask, &bitRegOffset);
         
         WDC_PciReadCfg8(device, (DWORD)OFFSET_VSEC+bitRegOffset, &byteValue);
         byteValue = (unsigned char)(bitValue ? (byteValue | bitMask) : (byteValue & ~bitMask));
         WDC_PciWriteCfg8(device, (DWORD)OFFSET_VSEC+bitRegOffset, byteValue);         
         
         break;
      default:
         break; // do nothing, the 5 bits above are the only writeable ones
   }
}

void ACL_PCIE_CONFIG::CVP_DRV_SetNumClks(WDC_DEVICE_HANDLE device, unsigned char numClks)
{
   if (numClks == 64) {
      WDC_PciWriteCfg8(device, OFFSET_VSEC+OFFSET_CVP_NUMCLKS, 0x0);
   } else if (numClks > 0 && numClks < 64) {
      WDC_PciWriteCfg8(device, OFFSET_VSEC+OFFSET_CVP_NUMCLKS, numClks);
   } else {
      // numClks is not a valid number!
   }
}

// Issue "dummy" writes to the HIP's memory to make the CB switch between CvP and internal clocks.
// The CB needs at least 244 125 MHz clock ticks, so we give it 64*4 = 256.
void ACL_PCIE_CONFIG::switch_clock(WDC_DEVICE_HANDLE device)
{
   const BYTE NUM_REG_WRITES = 244;
   const BYTE NUM_CVP_CLKS   = 1;
   const UINT32 VALUE_DUMMY  = 0x0;   

   BYTE i;
   CVP_DRV_SetNumClks(device, NUM_CVP_CLKS);
   for (i = 0; i < NUM_REG_WRITES; i++) {
      WDC_WriteAddr32(device, 0 /* BAR0 */, OFFSET_NONE, VALUE_DUMMY);
   }
}

void ACL_PCIE_CONFIG::prepare_for_pgm_data(WDC_DEVICE_HANDLE device)
{
   if (read_bit(device, DATA_COMPRESSED)) {
      CVP_DRV_SetNumClks(device, 8);
   } else if (read_bit(device, DATA_ENCRYPTED)) {
      CVP_DRV_SetNumClks(device, 4);
   } else {
      CVP_DRV_SetNumClks(device, 1);
   }
}

// check for CRC error that many 32-bit words
size_t ACL_PCIE_CONFIG::send_pgm_data(WDC_DEVICE_HANDLE device, unsigned int *data, size_t numWords)
{
   const size_t ERR_CHK_INTERVAL = 25000;
   size_t i;
   for (i = 0; i < numWords; i++){
      WDC_WriteAddr32(device, 0 /* BAR0 */, OFFSET_NONE, data[i]); 
      ACL_PCIE_ERROR_IF( (i % ERR_CHK_INTERVAL == 0) && read_bit(device, CVP_CONFIG_ERROR), break,
         "CB detected a CRC error between words %d and %d!\n", (i - ERR_CHK_INTERVAL), i);
   }
   return i; // return the number of Words sent
}

#endif   // WINDOWS

